@@ -465,10 +465,7 @@ void Val::Describe(ODesc* d) const
 		d->SP();
 		}
 
-	if ( d->IsReadable() )
-		ValDescribe(d);
-	else
-		Val::ValDescribe(d);
+	ValDescribe(d);
 	}
 
 void Val::DescribeReST(ODesc* d) const
@@ -2988,8 +2985,10 @@ bool VectorVal::Assign(unsigned int index, Val* element, Opcode op)
 			}
 		}
 
+	Val* val_at_index = 0;
+
 	if ( index < val.vector_val->size() )
-		Unref((*val.vector_val)[index]);
+		val_at_index = (*val.vector_val)[index];
 	else
 		val.vector_val->resize(index + 1);
 
@@ -3002,10 +3001,12 @@ bool VectorVal::Assign(unsigned int index, Val* element, Opcode op)
 
 		StateAccess::Log(new StateAccess(op == OP_INCR ?
 				OP_INCR_IDX : OP_ASSIGN_IDX,
-				this, ival, element, (*val.vector_val)[index]));
+				this, ival, element, val_at_index));
 		Unref(ival);
 		}
 
+	Unref(val_at_index);
+
 	// Note: we do *not* Ref() the element, if any, at this point.
 	// AssignExpr::Eval() already does this; other callers must remember
 	// to do it similarly.
