@@ -239,14 +239,19 @@ int HTTP_Entity::Undelivered(int64_t len)
 			  len, expect_data_length);
 		}
 
-	if ( end_of_data && in_header )
+	// Don't propogate an entity (file) gap if we're still in the headers,
+	// or the body length was declared to be zero.
+	if ( (end_of_data && in_header) || body_length == 0 )
 		return 0;
 
 	if ( is_partial_content )
+		{
 		precomputed_file_id = file_mgr->Gap(body_length, len,
 		              http_message->MyHTTP_Analyzer()->GetAnalyzerTag(),
 		              http_message->MyHTTP_Analyzer()->Conn(),
 		              http_message->IsOrig(), precomputed_file_id);
+		offset += len;
+		}
 	else
 		precomputed_file_id = file_mgr->Gap(body_length, len,
 		                  http_message->MyHTTP_Analyzer()->GetAnalyzerTag(),
@@ -413,7 +418,7 @@ void HTTP_Entity::SubmitHeader(mime::MIME_Header* h)
 		int64_t len = l - f + 1;
 
 		if ( DEBUG_http )
-			DEBUG_MSG("Content-Range length = %"PRId64"\n", len);
+			DEBUG_MSG("Content-Range length = %" PRId64"\n", len);
 
 		if ( len > 0 )
 			{
@@ -463,6 +468,20 @@ void HTTP_Entity::SubmitAllHeaders()
 	if ( DEBUG_http )
 		DEBUG_MSG("%.6f end of headers\n", network_time);
 
+	if ( Parent() &&
+	     Parent()->MIMEContentType() == mime::CONTENT_TYPE_MULTIPART )
+		{
+		// Don't treat single \r or \n characters in the multipart body content
+		// as lines because the MIME_Entity code will implicitly add back a
+		// \r\n for each line it receives.  We do this instead of setting
+		// plain delivery mode for the content line analyzer because
+		// the size of the content to deliver "plainly" may be unknown
+		// and just leaving it in that mode indefinitely screws up the
+		// detection of multipart boundaries.
+		http_message->content_line->SupressWeirds(true);
+		http_message->content_line->SetCRLFAsEOL(0);
+		}
+
 	// The presence of a message-body in a request is signaled by
 	// the inclusion of a Content-Length or Transfer-Encoding
 	// header field in the request's message-headers.
@@ -542,12 +561,9 @@ HTTP_Message::HTTP_Message(HTTP_Analyzer* arg_analyzer,
 
 	current_entity = 0;
 	top_level = new HTTP_Entity(this, 0, expect_body);
+	entity_data_buffer = 0;
 	BeginEntity(top_level);
 
-	buffer_offset = buffer_size = 0;
-	data_buffer = 0;
-	total_buffer_size = 0;
-
 	start_time = network_time;
 	body_length = 0;
 	content_gap_length = 0;
@@ -557,6 +573,7 @@ HTTP_Message::HTTP_Message(HTTP_Analyzer* arg_analyzer,
 HTTP_Message::~HTTP_Message()
 	{
 	delete top_level;
+	delete [] entity_data_buffer;
 	}
 
 Val* HTTP_Message::BuildMessageStat(const int interrupted, const char* msg)
@@ -604,22 +621,14 @@ void HTTP_Message::Done(const int interrupted, const char* detail)
 		}
 
 	MyHTTP_Analyzer()->HTTP_MessageDone(is_orig, this);
-
-	delete_strings(buffers);
-
-	if ( data_buffer )
-		{
-		delete data_buffer;
-		data_buffer = 0;
-		}
 	}
 
 int HTTP_Message::Undelivered(int64_t len)
 	{
-	if ( ! top_level )
-		return 0;
+	HTTP_Entity* e = current_entity ? current_entity
+	                                : static_cast<HTTP_Entity*>(top_level);
 
-	if ( ((HTTP_Entity*) top_level)->Undelivered(len) )
+	if ( e && e->Undelivered(len) )
 		{
 		content_gap_length += len;
 		return 1;
@@ -652,8 +661,6 @@ void HTTP_Message::EndEntity(mime::MIME_Entity* entity)
 	body_length += ((HTTP_Entity*) entity)->BodyLength();
 	header_length += ((HTTP_Entity*) entity)->HeaderLength();
 
-	DeliverEntityData();
-
 	if ( http_end_entity )
 		{
 		val_list* vl = new val_list();
@@ -664,6 +671,13 @@ void HTTP_Message::EndEntity(mime::MIME_Entity* entity)
 
 	current_entity = (HTTP_Entity*) entity->Parent();
 
+	if ( entity->Parent() &&
+	     entity->Parent()->MIMEContentType() == mime::CONTENT_TYPE_MULTIPART )
+		{
+		content_line->SupressWeirds(false);
+		content_line->SetCRLFAsEOL();
+		}
+
 	// It is necessary to call Done when EndEntity is triggered by
 	// SubmitAllHeaders (through EndOfData).
 	if ( entity == top_level )
@@ -720,31 +734,18 @@ void HTTP_Message::SubmitTrailingHeaders(mime::MIME_HeaderList& /* hlist */)
 
 void HTTP_Message::SubmitData(int len, const char* buf)
 	{
-	if ( buf != (const char*) data_buffer->Bytes() + buffer_offset ||
-	     buffer_offset + len > buffer_size )
-		{
-		reporter->AnalyzerError(MyHTTP_Analyzer(),
-		                                "HTTP message buffer misalignment");
-		return;
-		}
-
-	buffer_offset += len;
-	if ( buffer_offset >= buffer_size )
-		{
-		buffers.push_back(data_buffer);
-		data_buffer = 0;
-		}
+	if ( http_entity_data )
+		MyHTTP_Analyzer()->HTTP_EntityData(is_orig,
+		        new BroString(reinterpret_cast<const u_char*>(buf), len, 0));
 	}
 
 int HTTP_Message::RequestBuffer(int* plen, char** pbuf)
 	{
-	if ( ! data_buffer )
-		if ( ! InitBuffer(mime_segment_length) )
-			return 0;
-
-	*plen = data_buffer->Len() - buffer_offset;
-	*pbuf = (char*) data_buffer->Bytes() + buffer_offset;
+	if ( ! entity_data_buffer )
+		entity_data_buffer = new char[http_entity_data_delivery_size];
 
+	*plen = http_entity_data_delivery_size;
+	*pbuf = entity_data_buffer;
 	return 1;
 	}
 
@@ -785,9 +786,6 @@ void HTTP_Message::SetPlainDelivery(int64_t length)
 
 	if ( length > 0 && BifConst::skip_http_data )
 		content_line->SkipBytesAfterThisLine(length);
-
-	if ( ! data_buffer )
-		InitBuffer(length);
 	}
 
 void HTTP_Message::SkipEntityData()
@@ -796,87 +794,6 @@ void HTTP_Message::SkipEntityData()
 		current_entity->SkipBody();
 	}
 
-void HTTP_Message::DeliverEntityData()
-	{
-	if ( http_entity_data )
-		{
-		const BroString* entity_data = 0;
-
-		if ( data_buffer && buffer_offset > 0 )
-			{
-			if ( buffer_offset < buffer_size )
-				{
-				entity_data = new BroString(data_buffer->Bytes(), buffer_offset, 0);
-				delete data_buffer;
-				}
-			else
-				entity_data = data_buffer;
-
-			data_buffer = 0;
-
-			if ( buffers.empty() )
-				MyHTTP_Analyzer()->HTTP_EntityData(is_orig,
-								entity_data);
-			else
-				buffers.push_back(entity_data);
-
-			entity_data = 0;
-			}
-
-		if ( ! buffers.empty() )
-			{
-			if ( buffers.size() == 1 )
-				{
-				entity_data = buffers[0];
-				buffers.clear();
-				}
-			else
-				{
-				entity_data = concatenate(buffers);
-				delete_strings(buffers);
-				}
-
-			MyHTTP_Analyzer()->HTTP_EntityData(is_orig, entity_data);
-			}
-		}
-	else
-		{
-		delete_strings(buffers);
-
-		if ( data_buffer )
-			delete data_buffer;
-
-		data_buffer = 0;
-		}
-
-	total_buffer_size = 0;
-	}
-
-int HTTP_Message::InitBuffer(int64_t length)
-	{
-	if ( length <= 0 )
-		return 0;
-
-	if ( total_buffer_size >= http_entity_data_delivery_size )
-		DeliverEntityData();
-
-	if ( total_buffer_size + length > http_entity_data_delivery_size )
-		{
-		length = http_entity_data_delivery_size - total_buffer_size;
-		if ( length <= 0 )
-			return 0;
-		}
-
-	u_char* b = new u_char[length];
-	data_buffer = new BroString(0, b, length);
-
-	buffer_size = length;
-	total_buffer_size += length;
-	buffer_offset = 0;
-
-	return 1;
-	}
-
 void HTTP_Message::Weird(const char* msg)
 	{
 	analyzer->Weird(msg);
@@ -1070,9 +987,7 @@ void HTTP_Analyzer::DeliverStream(int len, const u_char* data, bool is_orig)
 				{
 				++num_replies;
 
-				if ( unanswered_requests.empty() )
-					Weird("unmatched_HTTP_reply");
-				else
+				if ( ! unanswered_requests.empty() )
 					ProtocolConfirmation();
 
 				reply_state = EXPECT_REPLY_MESSAGE;
@@ -1145,7 +1060,7 @@ void HTTP_Analyzer::Undelivered(uint64 seq, int len, bool is_orig)
 		{
 		if ( msg )
 			msg->SubmitEvent(mime::MIME_EVENT_CONTENT_GAP,
-				fmt("seq=%"PRIu64", len=%d", seq, len));
+				fmt("seq=%" PRIu64", len=%d", seq, len));
 		}
 
 	// Check if the content gap falls completely within a message body
@@ -1823,7 +1738,7 @@ void HTTP_Analyzer::ParseVersion(data_chunk_t ver, const IPAddr& host,
 		}
 	}
 
-void HTTP_Analyzer::HTTP_EntityData(int is_orig, const BroString* entity_data)
+void HTTP_Analyzer::HTTP_EntityData(int is_orig, BroString* entity_data)
 	{
 	if ( http_entity_data )
 		{
@@ -1831,8 +1746,7 @@ void HTTP_Analyzer::HTTP_EntityData(int is_orig, const BroString* entity_data)
 		vl->append(BuildConnVal());
 		vl->append(new Val(is_orig, TYPE_BOOL));
 		vl->append(new Val(entity_data->Len(), TYPE_COUNT));
-		// FIXME: Make sure that removing the const here is indeed ok...
-		vl->append(new StringVal(const_cast<BroString*>(entity_data)));
+		vl->append(new StringVal(entity_data));
 		ConnectionEvent(http_entity_data, vl);
 		}
 	else
